#include "ircd.h"
#include "addr.h"
#include "pest.h"
#include "peer.h"
#include "crypto.h"
#include "net.h"
#include "log.h"
#include "util.h"
#include "data.h"
#include "wipe.h"
#include "state.h"
#include "string.h"
#include <string.h>
#include <stdlib.h>

static void irc_sendmotd(void){
  ircd_sendf(":%s 375 %s :- %s Message of the Day -",
             IRC_SERV,self_handle,IRC_SERV);
  ircd_sendf(":%s 372 %s :- %s",
             IRC_SERV,self_handle,IRC_MOTD);
  ircd_sendf(":%s 376 %s :End of MOTD command.",
             IRC_SERV,self_handle);
}

static void irc_welcome(void){
  if( irc_ready  ||
     !irc_conned ||
     !irc_usered ||
     !irc_passed ||
     !irc_nicked) return;
  irc_ready = 1;

  ircd_sendf(":%s 001 %s :Welcome to pest %s!%s@%s",
             IRC_SERV,self_handle,self_handle,irc_user,irc_host);
  ircd_sendf(":%s 002 %s :Your host is %s, running version "
             PROG_FULLNAME,
             IRC_SERV,self_handle,IRC_SERV,PROG_VERSION);
  ircd_sendf(":%s 003 %s :This server was created at some point",
             IRC_SERV,self_handle);
  ircd_sendf(":%s 004 %s %s " PROG_FULLNAME " - -",
             IRC_SERV,self_handle,IRC_SERV,PROG_VERSION);
  irc_sendmotd();

  log_out("operator connected");
}

static void irc_names(char const *chan){
  if(strncmpci(chan,irc_chan,IRC_CHAN_LEN) == 0){
    char  names[(HANDLE_LEN+1)*(1+PEER_MAX)+1]; //"handle "*(self+peers)\0
    char *n = names;
    char *e = names+sizeof(names);

    *n  = 0;
    n   = strpcpy(n,self_handle,e);
    for(size_t x = 0;x < peer_cnt;x++){
      peer_t *p = peer+x;
      if(!p->active) continue;
      n = strpcpy(n," ",e);
      n = strpcpy(n,p->handle[0],e);
    }

    ircd_sendf(":%s 353 %s @ %s :%s",
               IRC_SERV,self_handle,chan,names);
  }
  ircd_sendf(":%s 366 %s %s :End of NAMES list.",
             IRC_SERV,self_handle,chan);
}

static void ircop_user(ircmsg_t const *m){
  if(irc_usered) return;
  if(m->argc < 1 || m->argc > 3){
    return;
  }

  char *user = m->argv[0];
  if(strncmp(user,irc_user,IRC_USER_LEN) != 0){
    log_err("username mismatch: '%s' != '%s'",user,irc_user);
    ircd_ckill();
    return;
  }

  irc_usered = 1;
  log_out("username accepted");
  irc_welcome();
}

static void ircop_pass(ircmsg_t const *m){
  char     *pwd;
  uint8_t   pass[SHA512_SIZE];

  if(irc_passed) return;
  if(m->argc != 1){
    return;
  }

  pwd = m->argv[0];
  sha512(pass,(uint8_t *)pwd,strlen(pwd));
  if(memcmp(pass,irc_pass,SHA512_SIZE) != 0){
    log_err("password mismatch");
    ircd_ckill();
    return;
  }

  irc_passed = 1;
  log_out("password accepted");
  irc_welcome();
}

static void ircop_nick(ircmsg_t const *m){
  if(m->argc != 1){
    irc_out("usage: NICK handle");
    return;
  }

  char *handle = m->argv[0];
  if(!str_valid(handle,1,HANDLE_LEN,HANDLE_CHARSET)){
    ircd_sendf(":%s 432 %s :Erroneus handle",IRC_SERV,handle);
    irclog_err("invalid handle: '%s'",handle);
    return;
  }

  if(peer_lookup_noerr(handle)){
    ircd_sendf(":%s 433 %s :Nickname is already in use",IRC_SERV,handle);
    irclog_err("handle '%s' already exists in WOT",handle);
    return;
  }

  ircd_sendf(":%s!%s@%s NICK :%s",self_handle,irc_user,irc_host,handle);
  strzcpy(self_handle,handle,sizeof(self_handle));
  irc_nicked = 1;
  log_out("handle set to '%s'",self_handle);
  irc_welcome();
}

static void ircop_join(ircmsg_t const *m){
  if(m->argc != 1){
    irc_out("usage: JOIN channel");
    return;
  }

  char *chan = m->argv[0];
  if(chan[0] != '#' || strlen(chan) > IRC_CHAN_LEN){
    irclog_err("invalid channel name: '%s'",chan);
    return;
  }

  if(irc_joined){
    ircd_sendf(":%s!%s@%s PART %s :",self_handle,irc_user,irc_host,irc_chan);
    log_out("operator parted %s",irc_chan);
  }
  irc_joined = 1;
  strzcpy(irc_chan,chan,sizeof(irc_chan));

  ircd_sendf(":%s!%s@%s JOIN %s",self_handle,irc_user,irc_host,irc_chan);
  irc_names(irc_chan);
  log_out("operator joined %s",chan);
}

static void ircop_part(ircmsg_t const *m){
  if(m->argc != 1){
    irc_out("usage: PART channel");
    return;
  }

  char *chan = m->argv[0];
  if(strncmpci(chan,irc_chan,IRC_CHAN_LEN) != 0){
    ircd_sendf(":%s 442 %s :You're not on that channel",IRC_SERV,self_handle);
    return;
  }

  ircd_sendf(":%s!%s@%s PART %s :",self_handle,irc_user,irc_host,irc_chan);
  log_out("operator parted %s",irc_chan);
  irc_joined  = 0;
  *irc_chan   = 0;
}

static void ircop_privmsg(ircmsg_t const *m){
  if(m->argc != 1 || !m->trail){
    irc_out("usage: PRIVMSG [channel | handle] :message");
    return;
  }

  char *chan  = m->argv[0];
  char *msg   = m->trail;

  if(chan[0] == '#'){
    if(strncmpci(chan,irc_chan,IRC_CHAN_LEN) != 0){
      ircd_sendf(":%s 404 %s :Cannot send to channel",IRC_SERV,chan);
      return;
    }
    pest_broadcast(msg);
  }else if(!pest_direct(chan,msg)){
    ircd_sendf(":%s 401 %s :No such nick",IRC_SERV,chan);
  }
}

static void ircop_names(ircmsg_t const *m){
  if(m->argc != 1){
    irc_out("usage: NAMES channel");
    return;
  }

  char *chan = m->argv[0];
  irc_names(chan);
}

static void ircop_version(ircmsg_t const *m){
  if(m->argc != 0){
    irc_out("usage: VERSION");
    return;
  }

  ircd_sendf(":%s 351 %s Pest-0x%02X. %s :" PROG_FULLNAME,
             IRC_SERV,self_handle,PEST_VERSION,IRC_SERV);
}

static void ircop_motd(ircmsg_t const *m){
  if(m->argc > 1){
    irc_out("usage: MOTD [target]");
    return;
  }

  irc_sendmotd();
}

static void ircop_ping(ircmsg_t const *m){
  if(m->argc != 1){
    irc_out("usage: PING server");
    return;
  }

  char *serv = m->argv[0]; (void)serv;
  ircd_sendf(":%s PONG %s :%s",IRC_SERV,IRC_SERV,IRC_SERV);
}

static void ircop_quit(ircmsg_t const *m){
  (void)m;
  log_out("operator disconnected");
  ircd_ckill();
}

static void ircop_wot(ircmsg_t const *m){
  if(m->argc > 1){
    irc_out("usage: WOT [handle]");
    return;
  }

  if(m->argc > 0){
    char   *handle  = m->argv[0];
    peer_t *p       = peer_lookup(handle);
    if(!p) return;
    peer_print(p);
  }else{
    peer_wot_print();
  }
}

static void ircop_peer(ircmsg_t const *m){
  char    host[IRC_HOST_BUF];
  peer_t *p;

  if(m->argc != 1){
    irc_out("usage: PEER handle");
    return;
  }

  char *handle = m->argv[0];
  p = peer_add(handle);
  if(!p) return;
  fmt_peerhost(host,sizeof(host),p->handle[0]);
  //ircd_sendf(":%s JOIN %s",host,irc_chan);
}

static void ircop_unpeer(ircmsg_t const *m){
  char host[IRC_HOST_BUF];

  if(m->argc != 1){
    irc_out("usage: UNPEER handle");
    return;
  }

  char   *handle  = m->argv[0];
  peer_t *p       = peer_lookup(handle);
  if(!p) return;
  fmt_peerhost(host,sizeof(host),p->handle[0]);
  if(p->active){
    ircd_sendf(":%s PART %s",host,irc_chan);
  }
  peer_del(p);
}

static void ircop_aka(ircmsg_t const *m){
  if(m->argc != 2){
    irc_out("usage: AKA handle alias");
    return;
  }

  char   *handle  = m->argv[0];
  char   *alias   = m->argv[1];
  peer_t *p       = peer_lookup(handle);
  if(!p) return;
  if(!peer_alias_add(p,alias)) return;
}

static void ircop_unaka(ircmsg_t const *m){
  char host[IRC_HOST_BUF];

  if(m->argc != 1){
    irc_out("usage: UNAKA handle");
    return;
  }

  char   *handle  = m->argv[0];
  peer_t *p       = peer_lookup(handle);
  if(!p) return;

  fmt_peerhost(host,sizeof(host),p->handle[0]);
  unsigned r = peer_alias_del(p,handle);
  if(!r) return;
  if(r == 2){ //TODO: this is disgusting but works
    ircd_sendf(":%s NICK :%s",host,p->handle[0]);
  }
}

static void ircop_pause(ircmsg_t const *m){
  if(m->argc != 1){
    irc_out("usage: PAUSE handle");
    return;
  }

  char   *handle  = m->argv[0];
  peer_t *p       = peer_lookup(handle);
  if(!p) return;
  peer_pause(p,1);
}

static void ircop_unpause(ircmsg_t const *m){
  if(m->argc != 1){
    irc_out("usage: UNPAUSE handle");
    return;
  }

  char   *handle  = m->argv[0];
  peer_t *p       = peer_lookup(handle);
  if(!p) return;
  peer_pause(p,0);
}

static void ircop_key(ircmsg_t const *m){
  if(m->argc != 2){
    irc_out("usage: KEY handle key");
    return;
  }

  uint8_t key[KEY_SIZE];
  char   *handle  = m->argv[0];
  char   *b64_key = m->argv[1];
  peer_t *p       = peer_lookup(handle);
  if(!p) return;

  if(!key_from_b64(key,b64_key)){
    irc_err("invalid key: '%s'",b64_key);
    key_zero(key);
    return;
  }

  peer_key_add(p,key,0);
  key_zero(key);
}

static void ircop_unkey(ircmsg_t const *m){
  uint8_t key[KEY_SIZE];

  if(m->argc != 1){
    irc_out("usage: UNKEY key");
    return;
  }

  char *b64_key = m->argv[0];
  if(!key_from_b64(key,b64_key)){
    irc_err("invalid key: '%s'",b64_key);
    key_zero(key);
    return;
  }

  peer_key_del(key,0);
  key_zero(key);
}

static void ircop_genkey(ircmsg_t const *m){
  uint8_t key     [KEY_SIZE];
  char    b64_key [B64_KEY_BUF];

  if(m->argc != 0){
    irc_out("usage: GENKEY");
    return;
  }

  if(!key_gen(key)){
    irclog_err("could not generate key");
    return;
  }

  b64_from_key(b64_key,sizeof(b64_key),key);
  irc_out("key generated from '%s': %s",RNG_PATH,b64_key);
  key_zero(key);
  BUF_ZERO(b64_key);
}

static void ircop_rekey(ircmsg_t const *m){
  if(m->argc > 1){
    irc_out("usage: REKEY [handle]");
    return;
  }

  if(m->argc == 0){
    rekey_initiate(0,1);
  }else{
    char   *handle  = m->argv[0];
    peer_t *p       = peer_lookup(handle);
    if(!p) return;
    rekey_initiate(p,1);
  }
}

static void ircop_rktog(ircmsg_t const *m){
  if(m->argc > 1){
  usage:
    irc_out("usage: RKTOG ['ENABLE'/'DISABLE']");
    return;
  }

  if(m->argc == 0){
    knob_display(FILE_RKRECV,0);
  }else if(strcmpci(m->argv[0],"ENABLE") == 0){
    knob_set(FILE_RKRECV,"1");
  }else if(strcmpci(m->argv[0],"DISABLE") == 0){
    knob_set(FILE_RKRECV,"0");
  }else{
    goto usage;
  }
}

static void ircop_gag(ircmsg_t const *m){
  if(m->argc > 1){
    irc_out("usage: GAG [handle]");
    return;
  }

  if(m->argc == 0){
    killfile_print();
  }else{
    char *handle = m->argv[0];
    killfile_add(handle);
  }
}

static void ircop_ungag(ircmsg_t const *m){
  if(m->argc != 1){
    irc_out("usage: UNGAG handle");
    return;
  }

  char *handle = m->argv[0];
  killfile_del(handle);
}

static void ircop_at(ircmsg_t const *m){
  if(m->argc > 2){
    irc_out("usage: AT [handle [ip:port]]");
    return;
  }

  if(m->argc == 0){
    peer_at_print();
  }else if(m->argc == 1){
    char   *handle  = m->argv[0];
    peer_t *p       = peer_lookup(handle);
    if(!p) return;
    if(addr_null(&p->addr)){
      irclog_out("no address stored for peer '%s'",handle);
      return;
    }
    peer_addr_print(p);
  }else{
    char   *handle  = m->argv[0];
    char   *at      = m->argv[1];
    peer_t *p       = peer_lookup(handle);
    addr_t  addr;

    if(!p) return;
    if(!addr_from_at(&addr,at)){
      irclog_err("invalid address (format is ip:port): '%s'",at);
      return;
    }

    //char host[IRC_HOST_BUF];
    //fmt_peerhost(host,sizeof(host),p->handle[0]);
    //ircd_sendf(":%s PART %s",host,irc_chan);
    //peer_addr_set(p,&addr);
    //fmt_peerhost(host,sizeof(host),p->handle[0]);
    //ircd_sendf(":%s JOIN %s",host,irc_chan);
    peer_addr_set(p,&addr,1);
  }
}

static void ircop_knob(ircmsg_t const *m){
  if((!m->trail && m->argc  > 2) ||
     ( m->trail && m->argc != 1)){
    irc_out("usage: KNOB [knob [[:]value]]");
    return;
  }

  if(m->argc == 0){
    knob_display(0,0);
  }else if(m->trail){
    char *knob  = m->argv[0];
    char *value = m->trail;
    knob_set(knob,value);
  }else if(m->argc == 1){
    char *knob  = m->argv[0];
    knob_display(knob,0);
  }else{
    char *knob  = m->argv[0];
    char *value = m->argv[1];
    knob_set(knob,value);
  }
}

static void ircop_knobinfo(ircmsg_t const *m){
  if(m->argc != 1){
    irc_out("usage: KNOBINFO knob");
    return;
  }

  char *knob = m->argv[0];
  knob_display(knob,1);
}

static void ircop_cut(ircmsg_t const *m){
  if(m->argc > 1){
    irc_out("usage: CUT [cutoff]");
    return;
  }

  if(m->argc == 0){
    knob_display(FILE_CUTOFF,0);
  }else{
    knob_set(FILE_CUTOFF,m->argv[0]);
  }
}

static void ircop_resolve(ircmsg_t const *m){
  if(m->argc != 1){
    irc_out("usage: RESOLVE handle");
    return;
  }

  char *handle = m->argv[0];
  pest_resolve(handle);
}

static void ircop_slave(ircmsg_t const *m){
  if(m->argc > 1){
    irc_out("usage: SLAVE [handle]");
    return;
  }

  if(!m->argc){
    if(*master){
      irc_out("slave mode is enabled (master = '%s')",master);
    }else{
      irc_out("slave mode is disabled");
    }
    return;
  }

  if(*master){
    irc_err("slave mode already enabled");
    return;
  }

  char *handle = m->argv[0];
  knob_set(FILE_MASTER,handle);
  irc_out("slave mode enabled");
}

static void ircop_unslave(ircmsg_t const *m){
  if(m->argc != 0){
    irc_out("usage: UNSLAVE");
    return;
  }

  if(!*master){
    irc_err("slave mode already disabled");
    return;
  }

  knob_set(FILE_MASTER,"");
  irc_out("slave mode disabled");
}

static void ircop_achtung(ircmsg_t const *m){
  if(m->argc != 0 || !m->trail){
    irc_out("usage: ACHTUNG :message");
    return;
  }

  char *msg = m->trail;
  pest_achtung(msg);
}

static void ircop_scram(ircmsg_t const *m){
  char     *pwd;
  uint8_t   pass[SHA512_SIZE];

  if(m->argc != 1){
    irc_out("usage: SCRAM scrampass");
    return;
  }

  pwd = m->argv[0];
  sha512(pass,(uint8_t *)pwd,strlen(pwd));
  if(memcmp(pass,scram_pass,SHA512_SIZE) != 0){
    irclog_err("SCRAM password mismatch");
    return;
  }

  scram_push();
}

#define IRC_OP(N)           { #N, 0, ircop_##N },
#define IRC_OP_INIT(N)      { #N, 1, ircop_##N },
#define IRC_OP_IGNORE(N)    { #N, 0, 0         },

typedef void (*ircmsg_f)(ircmsg_t const *);

typedef struct ircop{
  char const       *cmd;
  bool              init;
  ircmsg_f          func;
}ircop_t;

static ircop_t const irc_op[] = {
  IRC_OP_INIT   (user)
  IRC_OP_INIT   (pass)
  IRC_OP_INIT   (nick)
  IRC_OP        (join)
  IRC_OP        (part)
  IRC_OP        (privmsg)
  IRC_OP        (names)
  IRC_OP        (version)
  IRC_OP        (motd)
  IRC_OP        (ping)
  IRC_OP_IGNORE (cap)
  IRC_OP_IGNORE (mode)
  IRC_OP_IGNORE (notice)
  IRC_OP_IGNORE (who)
  IRC_OP_IGNORE (whois)
  IRC_OP_IGNORE (whowas)
  IRC_OP        (quit)

  IRC_OP        (wot)
  IRC_OP        (peer)
  IRC_OP        (unpeer)
  IRC_OP        (aka)
  IRC_OP        (unaka)
  IRC_OP        (pause)
  IRC_OP        (unpause)
  IRC_OP        (key)
  IRC_OP        (unkey)
  IRC_OP        (genkey)
  IRC_OP        (rekey)
  IRC_OP        (rktog)
  IRC_OP        (gag)
  IRC_OP        (ungag)
  IRC_OP        (at)
  IRC_OP        (knob)
  IRC_OP        (knobinfo)
  IRC_OP        (cut)
  IRC_OP        (resolve)
  IRC_OP        (slave)
  IRC_OP        (unslave)
  IRC_OP        (achtung)
  IRC_OP        (scram)
};

void irc_ev(unsigned e,ircmsg_t const *msg);
void irc_ev(unsigned e,ircmsg_t const *msg){
  switch(e){
    char cmd[IRCD_CMD_BUF];

  case IRCD_CONNECT:
    ip_str(irc_host,sizeof(irc_host),ircd_getip());
    irc_conned = 1;
    break;

  case IRCD_DISCONNECT:
    *irc_host   = 0;
    irc_conned  = 0;
    irc_usered  = 0;
    irc_passed  = 0;
    irc_nicked  = 0;
    irc_joined  = 0;
    irc_ready   = 0;
    break;

  case IRCD_MESSAGE:
    for(size_t x = 0;x < ARR_CNT(irc_op);x++){
      ircop_t const *op = irc_op+x;
      if(strncmp(msg->cmd,op->cmd,IRCD_CMD_BUF) != 0) continue;
      if(op->init || irc_ready){
        if(op->func) op->func(msg);
      }
      return;
    }

    strncpy(cmd,msg->cmd,sizeof(cmd));
    str_toupper(cmd);
    irclog_err("unknown/unhandled IRC command received: %s",cmd);
    break;

  default:
    assert(0);
  }
}

