#define _POSIX_C_SOURCE 200112L
#include "ircd.h"
#include "pest.h"
#include "peer.h"
#include "crypto.h"
#include "io.h"
#include "log.h"
#include "util.h"
#include "string.h"
#include "time.h"
#include "def.h"
#include "data.h"
#include "wipe.h"
#include "state.h"
#include <unistd.h>
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <signal.h>
#include <assert.h>

void irc_ev(unsigned e,ircmsg_t const *msg);

static int usage(int code){
  fputs(
    "usage: " PROG_NAME " [options]\n"
    "options:\n"
    "  -v           display version\n"
    "  -h           display usage\n"
    "  -u name      store username\n"
    "  -p pwd       store password\n"
    "  -s pwd       store SCRAM password\n"
    "  -i port      store IRC port\n"
    "  -q port      store pest port\n"
    "  -d path      data directory (default = '" DATA_PATH_DEF "')\n"
    "  -l path      log file (default = '" LOG_PATH_DEF "', "
                              "empty path = only stderr)\n"
    ,stderr
  );
  return code;
}

static void sigint_handler(int a){
  (void)a;
  running = 0;
}

static void irc_privmsg(char const *speaker,
                        char const *text,
                        char const *chan){

  char  buf[IRC_MSG_BUF];
  char *i = buf;
  char *e = buf+sizeof(buf);

  i = strpcpy(i,":",            e);
  i = strpcpy(i,speaker,        e);
  //i = strpcpy(i,"@",            e);
  //i = strpcpy(i,IRC_PEERHOST,   e);
  i = strpcpy(i," PRIVMSG ",    e);
  i = strpcpy(i,chan,           e);
  i = strpcpy(i," :",           e);
  i = strpcpy(i,text,           e);
  i = strpcpy(i,"\r\n",         e);

  ircd_send(buf);
  BUF_ZERO(buf);
}

static void irc_join(char const *speaker){
  char host[IRC_HOST_BUF];
  fmt_peerhost(host,sizeof(host),speaker);
  ircd_sendf(":%s JOIN %s",host,irc_chan);
}

static void irc_part(char const *speaker){
  char host[IRC_HOST_BUF];
  fmt_peerhost(host,sizeof(host),speaker);
  ircd_sendf(":%s PART %s",host,irc_chan);
}

static void pest_ev(unsigned e,char const *speaker,char const *text){
  if(!irc_ready) return;

  switch(e){
  case PEST_BROADCAST:
    if(irc_joined){
      irc_privmsg(speaker,text,irc_chan);
    }
    break;

  case PEST_DIRECT:
    irc_privmsg(speaker,text,self_handle);
    break;

  case PEST_ACTIVE:
    irc_join(speaker);
    break;

  case PEST_INACTIVE:
    irc_part(speaker);
    break;

  default:
    assert(0);
  }
}

static int _main(int argc,char **argv){
  int       opt;
  bool      store_user      = 0;
  bool      store_pass      = 0;
  bool      store_spass     = 0;
  bool      store_irc_port  = 0;
  bool      store_pest_port = 0;
  bool      loaded_user     = 0;
  bool      loaded_pass     = 0;
  uint64_t  time_start;
  uint64_t  time;
  uint64_t  ntime;

  while((opt = getopt(argc,argv,"vhu:p:s:i:q:d:l:")) != -1){
    switch(opt){
      int64_t port;

    case 'v':
      fprintf(stderr,PROG_FULLNAME "\n");
      return 0;

    case 'h':
      return usage(0);

    case 'u':
      if(!check_irc_user(optarg)) return 1;
      strzcpy(irc_user,optarg,sizeof(irc_user));
      store_user = 1;
      break;

    case 'p':
      sha512(irc_pass,(uint8_t *)optarg,strlen(optarg));
      store_pass = 1;
      break;

    case 's':
      sha512(scram_pass,(uint8_t *)optarg,strlen(optarg));
      store_spass = 1;
      break;

    case 'i':
      port = strtoll(optarg,0,10);
      if(!check_port(port,"IRC")) return 1;
      irc_port       = port;
      store_irc_port = 1;
      break;

    case 'q':
      port = strtoll(optarg,0,10);
      if(!check_port(port,"pest")) return 1;
      pest_port       = port;
      store_pest_port = 1;
      break;

    case 'd':
      strzcpy(data_path,optarg,sizeof(data_path));
      break;

    case 'l':
      strzcpy(log_path,optarg,sizeof(log_path));
      break;

    case '?':
    default:
      return usage(1);
    }
  }

  argc -= optind;
  argv += optind;

  if(*log_path){
    if(!log_open(log_path)) return 1;
  }

  umask(DATA_UMASK);
  if(!mkdir_p(data_path)) return 1;

  if(store_user       && !irc_user_store())   return 1;
  if(store_pass       && !irc_pass_store())   return 1;
  if(store_irc_port   && !irc_port_store())   return 1;
  if(store_pest_port  && !pest_port_store())  return 1;
  if(store_spass      && !scram_pass_store()) return 1;

  loaded_user = irc_user_load();
  loaded_pass = irc_pass_load();
  if(!loaded_user)        return 1;
  if(!loaded_pass)        return 1;

  if(!irc_port_load())    return 1;
  if(!pest_port_load())   return 1;
  if(!scram_pass_load())  return 1;
  if(!killfile_load())    return 1;
  if(!peer_load())        return 1;

  signal(SIGINT,sigint_handler);
  signal(SIGPIPE,SIG_IGN);

  if(!rng_open(RNG_PATH)) return 1;
  ircd_init(irc_ev);
  if(!pest_init(pest_ev)) return 1;
  log_out(PROG_FULLNAME " station running");

  time_start = mono_sec();
  while(running){
    bool active;

    time    = mono_sec();
    ntime   = mono_nsec();
    active  = irc_joined || time-time_start > PEST_WAIT_IRC;

    ircd_tick();
    pest_tick(time,ntime,active);
  }

  log_out("station shutting down...");
  ircd_exit();
  if(!scram) rng_close();
  log_close();
  return 0;
}

int main(int argc,char **argv){
  //crash early if STACK_WIPE_SIZE set too high
  stack_test();

  int r = _main(argc,argv);

  if(scram){
    dir_zero(data_path);
    dir_rand(data_path);
    rm_rf_nolog(data_path);
    if(*log_path){
      file_zero(log_path);
      file_rand(log_path);
      rm_nolog(log_path);
    }
    state_zero();
    stack_zero();
    state_rand();
    stack_rand();
    rng_close();
  }else{
    state_zero();
    stack_zero();
  }

  return r;
}

